/* defects_spti.c   gcc -Wall -O2 -s -odefects_spti.exe defects_spti.c
 *
 * Program for reading disc defect data via SPTI, v0.1
 * Author: Christian Koch <mailto:christian_koch@gmx.de>
 *
 * Derived from SPTI example from Microsoft Windows DDK
 *
 * Usage of the works is permitted provided that this
 * instrument is retained with the works, so that any entity
 * that uses the works is notified of this instrument.
 *
 * DISCLAIMER: THE WORKS ARE WITHOUT WARRANTY.
 *
 * [2004, Fair License; rhid.com/fair]
 *
 * Changelog:
 *
 * 2004-05-22   defects_spti.c v0.1
 */



#include <windows.h>
#include <ddk/ntddscsi.h>
#include <ddk/scsi.h>
#include <stdio.h>



typedef struct _SCSI_PASS_THROUGH_WITH_BUFFERS
{
  SCSI_PASS_THROUGH spt;
  DWORD             Filler;            // realign buffers to double word boundary
  union
  {
    BYTE            bySenseBuf[SENSE_BUFFER_SIZE];
    SENSE_DATA      SenseData;
  };
  BYTE              byDataBuf[16384];
} SCSI_PASS_THROUGH_WITH_BUFFERS;

typedef struct _DEFECT_DESCRIPTOR  // used by structure DEFECT_LIST_HEADER
{
  DWORD             dwTrackAddress : 24;
  DWORD             dwHeadAddress : 8;
  DWORD             dwSectorAddress;
} DEFECT_DESCRIPTOR;

typedef struct _DEFECT_LIST_HEADER  // returned from SCSIOP_READ_DEFECT_DATA
{
  BYTE              byReserved;
  BYTE              byDefectListFormat;
  WORD              wDefectListLength;
  DEFECT_DESCRIPTOR DefectDescriptor[];
} DEFECT_LIST_HEADER;

void
PrintError(DWORD);

void
PrintInquiryData(BYTE const *);

BOOL
PrintStatusResults(BOOL, SCSI_PASS_THROUGH_WITH_BUFFERS const *);

void
PrintSenseInfo(SCSI_PASS_THROUGH_WITH_BUFFERS const *);

void
PrintDefectData(DEFECT_LIST_HEADER const *, DWORD);

WORD
wSwap16(WORD);

DWORD
dwSwap24(DWORD);

DWORD
dwSwap32(DWORD);

//
// Command Descriptor Block constants
//

#define CDB6GENERIC_LENGTH                   6
#define CDB10GENERIC_LENGTH                  10

#define SCSIOP_READ_DEFECT_DATA              0x37



int
main(int argc, char *argv[])
{
  BOOL bStatus;
  DWORD dwAccessMode, dwShareMode;
  HANDLE hFile;
  SCSI_PASS_THROUGH_WITH_BUFFERS sptwb;
  BYTE byBuffer[2048];
  BYTE byString[25], byString2[15];
  DWORD dwLength, dwErrorCode, dwReturned;

  printf("Program for reading disc defect data via SPTI, v0.1\n");
  printf("Author: Christian Koch <mailto:christian_koch@gmx.de>\n\n");

  if (argc != 2)
  {
    printf("Usage:   %s <port-name>\n", argv[0]);
    printf("Example: %s d:\n\n", argv[0]);
    printf("    This program is distributed in the hope that it will be useful,\n");
    printf("    but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
    printf("    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");

    return 0;
  }

  lstrcpy(byString, "\\\\.\\");
  lstrcpyn(byString2, argv[1], sizeof(byString2));
  lstrcat(byString, byString2);

  dwShareMode = FILE_SHARE_READ | FILE_SHARE_WRITE;
  dwAccessMode = GENERIC_READ | GENERIC_WRITE;
  hFile = CreateFile(byString, dwAccessMode, dwShareMode, NULL, OPEN_EXISTING, 0, NULL);

  if (hFile == INVALID_HANDLE_VALUE)
  {
    printf("Error opening %s. Error: %lu\n", byString2, dwErrorCode = GetLastError());
    PrintError(dwErrorCode);
    return 1;
  }

  //
  // INQUIRY
  //

  printf("*** INQUIRY ***\n\n");

  ZeroMemory(&sptwb, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));

  bStatus = DeviceIoControl(hFile, IOCTL_SCSI_GET_INQUIRY_DATA, NULL, 0, byBuffer, sizeof(byBuffer), &dwReturned, FALSE);

  if (!PrintStatusResults(bStatus, &sptwb))
  {
    return 1;
  }

  PrintInquiryData(byBuffer);

  //
  // TEST UNIT READY
  //

  printf("*** TEST UNIT READY ***\n\n");

  ZeroMemory(&sptwb, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));
  sptwb.spt.Length = sizeof(SCSI_PASS_THROUGH);
  sptwb.spt.PathId = 0;
  sptwb.spt.TargetId = 1;
  sptwb.spt.Lun = 0;
  sptwb.spt.CdbLength = CDB6GENERIC_LENGTH;
  sptwb.spt.SenseInfoLength = SENSE_BUFFER_SIZE;
  sptwb.spt.DataIn = SCSI_IOCTL_DATA_IN;
  sptwb.spt.DataTransferLength = 0;
  sptwb.spt.TimeOutValue = 2;
  sptwb.spt.DataBufferOffset = FIELD_OFFSET(SCSI_PASS_THROUGH_WITH_BUFFERS, byDataBuf);
  sptwb.spt.SenseInfoOffset  = FIELD_OFFSET(SCSI_PASS_THROUGH_WITH_BUFFERS, bySenseBuf);
  sptwb.spt.Cdb[0] = SCSIOP_TEST_UNIT_READY;

  dwLength = FIELD_OFFSET(SCSI_PASS_THROUGH_WITH_BUFFERS, byDataBuf);

  bStatus = DeviceIoControl(hFile, IOCTL_SCSI_PASS_THROUGH, &sptwb, sizeof(SCSI_PASS_THROUGH), &sptwb, dwLength, &dwReturned, FALSE);

  if (!PrintStatusResults(bStatus, &sptwb))
  {
    return 1;
  }

  //
  // READ DEFECT DATA
  //

  printf("*** Primary Defect List ***\n\n");

  ZeroMemory(&sptwb, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));
  sptwb.spt.Length = sizeof(SCSI_PASS_THROUGH);
  sptwb.spt.PathId = 0;
  sptwb.spt.TargetId = 1;
  sptwb.spt.Lun = 0;
  sptwb.spt.CdbLength = CDB10GENERIC_LENGTH;
  sptwb.spt.SenseInfoLength = SENSE_BUFFER_SIZE;
  sptwb.spt.DataIn = SCSI_IOCTL_DATA_IN;
  sptwb.spt.DataTransferLength = sizeof(sptwb.byDataBuf);
  sptwb.spt.TimeOutValue = 2;
  sptwb.spt.DataBufferOffset = FIELD_OFFSET(SCSI_PASS_THROUGH_WITH_BUFFERS, byDataBuf);
  sptwb.spt.SenseInfoOffset  = FIELD_OFFSET(SCSI_PASS_THROUGH_WITH_BUFFERS, bySenseBuf);
  sptwb.spt.Cdb[0] = SCSIOP_READ_DEFECT_DATA;
  sptwb.spt.Cdb[2] = 0x15; // PDL
  sptwb.spt.Cdb[7] = HIBYTE(sizeof(sptwb.byDataBuf));
  sptwb.spt.Cdb[8] = LOBYTE(sizeof(sptwb.byDataBuf));

  dwLength = FIELD_OFFSET(SCSI_PASS_THROUGH_WITH_BUFFERS, byDataBuf) + sptwb.spt.DataTransferLength;

  bStatus = DeviceIoControl(hFile, IOCTL_SCSI_PASS_THROUGH, &sptwb, sizeof(SCSI_PASS_THROUGH), &sptwb, dwLength, &dwReturned, FALSE);

  if (PrintStatusResults(bStatus, &sptwb))
  {
    PrintDefectData((DEFECT_LIST_HEADER *) sptwb.byDataBuf, sptwb.spt.DataTransferLength);
  }

  printf("*** Secondary Defect List ***\n\n");

  ZeroMemory(&sptwb, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));
  sptwb.spt.Length = sizeof(SCSI_PASS_THROUGH);
  sptwb.spt.PathId = 0;
  sptwb.spt.TargetId = 1;
  sptwb.spt.Lun = 0;
  sptwb.spt.CdbLength = CDB10GENERIC_LENGTH;
  sptwb.spt.SenseInfoLength = SENSE_BUFFER_SIZE;
  sptwb.spt.DataIn = SCSI_IOCTL_DATA_IN;
  sptwb.spt.DataTransferLength = sizeof(sptwb.byDataBuf);
  sptwb.spt.TimeOutValue = 2;
  sptwb.spt.DataBufferOffset = FIELD_OFFSET(SCSI_PASS_THROUGH_WITH_BUFFERS, byDataBuf);
  sptwb.spt.SenseInfoOffset  = FIELD_OFFSET(SCSI_PASS_THROUGH_WITH_BUFFERS, bySenseBuf);
  sptwb.spt.Cdb[0] = SCSIOP_READ_DEFECT_DATA;
  sptwb.spt.Cdb[2] = 0x0D; // SDL
  sptwb.spt.Cdb[7] = HIBYTE(sizeof(sptwb.byDataBuf));
  sptwb.spt.Cdb[8] = LOBYTE(sizeof(sptwb.byDataBuf));

  dwLength = FIELD_OFFSET(SCSI_PASS_THROUGH_WITH_BUFFERS, byDataBuf) + sptwb.spt.DataTransferLength;

  bStatus = DeviceIoControl(hFile, IOCTL_SCSI_PASS_THROUGH, &sptwb, sizeof(SCSI_PASS_THROUGH), &sptwb, dwLength, &dwReturned, FALSE);

  if (PrintStatusResults(bStatus, &sptwb))
  {
    PrintDefectData((DEFECT_LIST_HEADER *) sptwb.byDataBuf, sptwb.spt.DataTransferLength);
  }

  printf("*** Primary and Secondary Defect List ***\n\n");

  ZeroMemory(&sptwb, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));
  sptwb.spt.Length = sizeof(SCSI_PASS_THROUGH);
  sptwb.spt.PathId = 0;
  sptwb.spt.TargetId = 1;
  sptwb.spt.Lun = 0;
  sptwb.spt.CdbLength = CDB10GENERIC_LENGTH;
  sptwb.spt.SenseInfoLength = SENSE_BUFFER_SIZE;
  sptwb.spt.DataIn = SCSI_IOCTL_DATA_IN;
  sptwb.spt.DataTransferLength = sizeof(sptwb.byDataBuf);
  sptwb.spt.TimeOutValue = 2;
  sptwb.spt.DataBufferOffset = FIELD_OFFSET(SCSI_PASS_THROUGH_WITH_BUFFERS, byDataBuf);
  sptwb.spt.SenseInfoOffset  = FIELD_OFFSET(SCSI_PASS_THROUGH_WITH_BUFFERS, bySenseBuf);
  sptwb.spt.Cdb[0] = SCSIOP_READ_DEFECT_DATA;
  sptwb.spt.Cdb[2] = 0x1D; // PDF + SDL
  sptwb.spt.Cdb[7] = HIBYTE(sizeof(sptwb.byDataBuf));
  sptwb.spt.Cdb[8] = LOBYTE(sizeof(sptwb.byDataBuf));

  dwLength = FIELD_OFFSET(SCSI_PASS_THROUGH_WITH_BUFFERS, byDataBuf) + sptwb.spt.DataTransferLength;

  bStatus = DeviceIoControl(hFile, IOCTL_SCSI_PASS_THROUGH, &sptwb, sizeof(SCSI_PASS_THROUGH), &sptwb, dwLength, &dwReturned, FALSE);

  if (PrintStatusResults(bStatus, &sptwb))
  {
    PrintDefectData((DEFECT_LIST_HEADER *) sptwb.byDataBuf, sptwb.spt.DataTransferLength);
  }

  return 0;
}



void
PrintError(DWORD dwErrorCode)
{
  BYTE byErrorBuffer[80];
  DWORD dwCount;

  dwCount = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, dwErrorCode, 0, byErrorBuffer, sizeof(byErrorBuffer), NULL);

  if (dwCount != 0)
  {
    printf("%s\n", byErrorBuffer);
  }
  else
  {
    printf("Format message failed.  Error: %lu\n", GetLastError());
  }
}



void
PrintInquiryData(BYTE const *pDataBuffer)
{
  SCSI_ADAPTER_BUS_INFO *pAdapterInfo;
  SCSI_INQUIRY_DATA *pInquiryData;
  int i, j;

  pAdapterInfo = (SCSI_ADAPTER_BUS_INFO *) pDataBuffer;

  printf("Bus\n");
  printf("Num TID LUN Claimed String                       Inquiry Header\n");
  printf("--- --- --- ------- ---------------------------- -----------------------\n");

  for (i = 0; i < pAdapterInfo->NumberOfBuses; i++)
  {
    pInquiryData = (SCSI_INQUIRY_DATA *) (pDataBuffer + pAdapterInfo->BusData[i].InquiryDataOffset);

    while (pAdapterInfo->BusData[i].InquiryDataOffset)
    {
      printf(" %d   %d  %3d    %s    %.28s ", i, pInquiryData->TargetId, pInquiryData->Lun, pInquiryData->DeviceClaimed ? "Y" : "N", &pInquiryData->InquiryData[8]);

      for (j = 0; j < 8; j++)
      {
        printf("%02X ", pInquiryData->InquiryData[j]);
      }

      printf("\n");

      if (pInquiryData->NextInquiryDataOffset == 0)
      {
        break;
      }

      pInquiryData = (SCSI_INQUIRY_DATA *) (pDataBuffer + pInquiryData->NextInquiryDataOffset);
    }
  }

  printf("\n\n");
}



BOOL
PrintStatusResults(BOOL bStatus, SCSI_PASS_THROUGH_WITH_BUFFERS const *psptwb)
{
  DWORD dwErrorCode;

  if (!bStatus )
  {
    printf( "Error: %lu  ", dwErrorCode = GetLastError() );
    PrintError(dwErrorCode);
    return FALSE;
  }

  if (psptwb->spt.ScsiStatus)
  {
     printf("SCSI status: %02Xh\n", psptwb->spt.ScsiStatus);

     if (psptwb->spt.SenseInfoLength >= SENSE_BUFFER_SIZE)
     {
       printf("ASC/ASCQ:    %02Xh/%02Xh\n", psptwb->SenseData.AdditionalSenseCode, psptwb->SenseData.AdditionalSenseCodeQualifier);
     }

     printf("\n");
     return FALSE;
  }

  return TRUE;
}



void
PrintDefectData(DEFECT_LIST_HEADER const * pDLH, DWORD dwLength)
{
  int i;
  DEFECT_DESCRIPTOR const *pDD;

  printf("   Defects: %hu\n\n", wSwap16(pDLH->wDefectListLength)/8);

  if ((wSwap16(pDLH->wDefectListLength)/8) && ((dwLength-4)/8))
  {
    printf("      #   Track   Head   Sector\n");
    printf("   ----------------------------\n");

    for (i = 0; i < min(wSwap16(pDLH->wDefectListLength)/8, (dwLength-4)/8); i++)
    {
      pDD = pDLH->DefectDescriptor + i;
      printf("   %4d   %5lu      %1d       %2lu\n", i+1 , dwSwap24(pDD->dwTrackAddress), pDD->dwHeadAddress, dwSwap32(pDD->dwSectorAddress));
    }

    printf("\n");
  }

  printf("\n");
}



WORD
wSwap16(WORD word)
{
  return((word << 8) | (word >> 8));
}



DWORD
dwSwap24(DWORD dword)
{
  return(((dword << 16) & 0x00FF0000) | (dword & 0x0000FF00) | ((dword >> 16) & 0x000000FF));
}



DWORD
dwSwap32(DWORD dword)
{
  return((dword << 24) | ((dword << 8) & 0x00FF0000) | ((dword >> 8) & 0x0000FF00) | (dword >> 24));
}
